import json

from sqlalchemy_utils import UUIDType
from sqlalchemy.types import PickleType

from director.extensions import db
from director.models import BaseModel, StatusType
from director.models.utils import JSONBType


class Task(BaseModel):
    __tablename__ = "tasks"

    key = db.Column(db.String(255), nullable=False)
    status = db.Column(db.Enum(StatusType), default=StatusType.pending, nullable=False)
    previous = db.Column(JSONBType, default=[])
    result = db.Column(PickleType)

    # Relationship
    workflow_id = db.Column(
        UUIDType(binary=False),
        db.ForeignKey("workflows.id"),
        nullable=False,
        index=True,
    )
    workflow = db.relationship("Workflow", backref=db.backref("tasks", lazy=True))

    def __repr__(self):
        return f"<Task {self.key}>"

    def to_dict(self):
        d = super().to_dict()
        d.update(
            {
                "key": self.key,
                "status": self.status.value,
                "task": self.id,
                "previous": self.previous,
                "result": self.result,
            }
        )
        return d



"""Add users table

Revision ID: 3f8466b16023
Revises: 05cf96d6fcae
Create Date: 2020-07-15 16:45:02.209593

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy_utils.types import UUIDType


# revision identifiers, used by Alembic.
revision = "3f8466b16023"
down_revision = "05cf96d6fcae"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "users",
        sa.Column("id", UUIDType(binary=False), nullable=False),
        sa.Column("created_at", sa.DateTime(timezone=True), nullable=False),
        sa.Column("updated_at", sa.DateTime(timezone=True), nullable=False),
        sa.Column("username", sa.String(255), nullable=False),
        sa.Column("password", sa.String(255), nullable=False),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_users")),
        sa.UniqueConstraint("username", name=op.f("uq_users_username")),
    )
    op.create_index(op.f("ix_users_created_at"), "users", ["created_at"], unique=False)


def downgrade():
    op.drop_index(op.f("ix_users_created_at"), table_name="users")
    op.drop_table("users")



from flask_json_schema import JsonValidationError
from jsonschema.validators import validator_for
from celery.schedules import crontab

from director.exceptions import WorkflowSyntaxError


def validate(payload, schema):
    """Validate a payload according to a given schema"""
    validator_cls = validator_for(schema)
    validator = validator_cls(schema=schema)
    errors = list(validator.iter_errors(payload))
    if errors:
        raise JsonValidationError("Payload is not valid", errors)


def format_schema_errors(e):
    """Format FlaskJsonSchema validation errors"""
    return {
        "error": e.message,
        "errors": [validation_err.message for validation_err in e.errors],
    }


def build_celery_schedule(workflow_name, data):
    """ A celery schedule can accept seconds or crontab """
    try:
        schedule = float(data)
    except ValueError:
        try:
            m, h, dw, dm, my = data.split(" ")

            schedule = crontab(
                minute=m,
                hour=h,
                day_of_week=dw,
                day_of_month=dm,
                month_of_year=my,
            )
        except Exception as e:
            raise WorkflowSyntaxError(workflow_name)

    return schedule



import click

from director import __version__
from director.commands.assets import dlassets
from director.commands.celery import celery
from director.commands.db import db
from director.commands.init import init
from director.commands.user import user
from director.commands.webserver import webserver
from director.commands.workflows import workflow


@click.version_option(version=__version__)
@click.group()
def cli():
    """Celery Director - Command Line Interface"""
    pass


cli.add_command(webserver)
cli.add_command(celery)
cli.add_command(workflow)
cli.add_command(db)
cli.add_command(user)
cli.add_command(dlassets)
cli.add_command(init)



from uuid import UUID

from flask import Blueprint, abort, render_template


view_bp = Blueprint("views", __name__, url_prefix="/")


@view_bp.route("/")
def home():
    return render_template("index.html")


@view_bp.route("/<id>")
def get_workflow(id):
    try:
        _ = UUID(id)
    except ValueError:
        abort(404)
    return render_template("index.html")


@view_bp.app_template_filter("status")
def status(code):
    return {
        "pending": "secondary",
        "progress": "warning",
        "success": "success",
        "error": "danger",
        "canceled": "dark",
    }[code]



from flask import current_app as app
from flask import abort, jsonify, request
from jsonschema.exceptions import ValidationError, SchemaError

from director.api import api_bp, validate
from director.auth import auth
from director.builder import WorkflowBuilder
from director.exceptions import WorkflowNotFound
from director.extensions import cel_workflows, schema
from director.models.workflows import Workflow


def _get_workflow(workflow_id):
    workflow = Workflow.query.filter_by(id=workflow_id).first()
    if not workflow:
        abort(404, f"Workflow {workflow_id} not found")
    return workflow


def _execute_workflow(project, name, payload={}):
    fullname = f"{project}.{name}"

    # Check if the workflow exists
    try:
        wf = cel_workflows.get_by_name(fullname)
        if "schema" in wf:
            validate(payload, wf["schema"])
    except WorkflowNotFound:
        abort(404, f"Workflow {fullname} not found")

    # Create the workflow in DB
    obj = Workflow(project=project, name=name, payload=payload)
    obj.save()

    # Build the workflow and execute it
    data = obj.to_dict()
    workflow = WorkflowBuilder(obj.id)
    workflow.run()

    app.logger.info(f"Workflow sent : {workflow.canvas}")
    return obj.to_dict(), workflow


@api_bp.route("/workflows", methods=["POST"])
@auth.login_required
@schema.validate(
    {
        "required": ["project", "name", "payload"],
        "additionalProperties": False,
        "properties": {
            "project": {"type": "string"},
            "name": {"type": "string"},
            "payload": {"type": "object"},
        },
    }
)
def create_workflow():
    project, name, payload = (
        request.get_json()["project"],
        request.get_json()["name"],
        request.get_json()["payload"],
    )
    data, _ = _execute_workflow(project, name, payload)
    return jsonify(data), 201


@api_bp.route("/workflows/<workflow_id>/relaunch", methods=["POST"])
@auth.login_required
def relaunch_workflow(workflow_id):
    obj = _get_workflow(workflow_id)
    data, _ = _execute_workflow(obj.project, obj.name, obj.payload)
    return jsonify(data), 201


@api_bp.route("/workflows")
@auth.login_required
def list_workflows():
    page = request.args.get("page", type=int, default=1)
    per_page = request.args.get(
        "per_page", type=int, default=app.config["WORKFLOWS_PER_PAGE"]
    )
    workflows = Workflow.query.order_by(Workflow.created_at.desc()).paginate(
        page, per_page
    )
    return jsonify([w.to_dict() for w in workflows.items])


@api_bp.route("/workflows/<workflow_id>")
@auth.login_required
def get_workflow(workflow_id):
    workflow = _get_workflow(workflow_id)
    tasks = [t.to_dict() for t in workflow.tasks]

    resp = workflow.to_dict()
    resp.update({"tasks": tasks})
    return jsonify(resp)



from uuid import UUID

import pytest

from director.exceptions import WorkflowNotFound
from director.tasks.workflows import start, end
from director.models.tasks import Task
from director.models.workflows import Workflow

from tests.conftest import _remove_keys


def test_create_unknown_workflow(create_builder):
    with pytest.raises(WorkflowNotFound):
        create_builder("project", "UNKNOW_WORKFLOW", {})


def test_build_one_task(create_builder):
    data, builder = create_builder("example", "WORKFLOW", {"foo": "bar"})
    assert data == {
        "name": "WORKFLOW",
        "payload": {"foo": "bar"},
        "project": "example",
        "fullname": "example.WORKFLOW",
        "status": "pending",
        "periodic": False,
    }

    assert len(builder.canvas) == 3
    assert builder.canvas[0].task == "director.tasks.workflows.start"
    assert builder.canvas[-1].task == "director.tasks.workflows.end"
    assert builder.canvas[1].task == "TASK_EXAMPLE"


def test_build_chained_tasks(app, create_builder):
    keys = ["id", "created", "updated", "task"]
    data, builder = create_builder("example", "SIMPLE_CHAIN", {"foo": "bar"})
    assert data == {
        "name": "SIMPLE_CHAIN",
        "payload": {"foo": "bar"},
        "project": "example",
        "fullname": "example.SIMPLE_CHAIN",
        "status": "pending",
        "periodic": False,
    }

    # Check the Celery canvas
    assert len(builder.canvas) == 5
    assert builder.canvas[0].task == "director.tasks.workflows.start"
    assert builder.canvas[-1].task == "director.tasks.workflows.end"
    assert builder.canvas[1].task == "TASK_A"
    assert builder.canvas[2].task == "TASK_B"
    assert builder.canvas[3].task == "TASK_C"

    # Check the tasks in database (including previouses ID)
    with app.app_context():
        tasks = Task.query.order_by(Task.created_at.asc()).all()
    assert len(tasks) == 3
    assert _remove_keys(tasks[0].to_dict(), keys) == {
        "key": "TASK_A",
        "previous": [],
        "result": None,
        "status": "pending",
    }
    assert _remove_keys(tasks[1].to_dict(), keys) == {
        "key": "TASK_B",
        "previous": [str(tasks[0].to_dict()["id"])],
        "result": None,
        "status": "pending",
    }

    assert _remove_keys(tasks[2].to_dict(), keys) == {
        "key": "TASK_C",
        "previous": [str(tasks[1].to_dict()["id"])],
        "result": None,
        "status": "pending",
    }


def test_build_grouped_tasks(app, create_builder):
    keys = ["id", "created", "updated", "task"]
    data, builder = create_builder("example", "SIMPLE_GROUP", {"foo": "bar"})
    assert data == {
        "name": "SIMPLE_GROUP",
        "payload": {"foo": "bar"},
        "project": "example",
        "fullname": "example.SIMPLE_GROUP",
        "status": "pending",
        "periodic": False,
    }

    # Check the Celery canvas
    assert len(builder.canvas) == 4
    assert builder.canvas[0].task == "director.tasks.workflows.start"
    assert builder.canvas[-1].task == "director.tasks.workflows.end"
    assert builder.canvas[1].task == "TASK_A"
    assert builder.canvas[2].task == "celery.group"
    group_tasks = builder.canvas[2].tasks
    assert len(group_tasks) == 2
    assert [group_tasks[0].task, group_tasks[1].task] == [
        "TASK_B",
        "TASK_C",
    ]

    # Check the tasks in database (including previouses ID)
    with app.app_context():
        tasks = Task.query.order_by(Task.created_at.asc()).all()
    assert len(tasks) == 3
    assert _remove_keys(tasks[0].to_dict(), keys) == {
        "key": "TASK_A",
        "previous": [],
        "result": None,
        "status": "pending",
    }
    assert _remove_keys(tasks[1].to_dict(), keys) == {
        "key": "TASK_B",
        "previous": [str(tasks[0].to_dict()["id"])],
        "result": None,
        "status": "pending",
    }
    assert _remove_keys(tasks[2].to_dict(), keys) == {
        "key": "TASK_C",
        "previous": [str(tasks[0].to_dict()["id"])],
        "result": None,
        "status": "pending",
    }



from pathlib import Path

import pytest
from celery.canvas import _chain
from celery.contrib.testing import worker
from flask import Response
from flask.testing import FlaskClient

from director import create_app
from director.builder import WorkflowBuilder
from director.extensions import cel, db
from director.models.workflows import Workflow


KEYS_TO_REMOVE = ["id", "created", "updated"]


class DirectorTestClient(FlaskClient):
    def __init__(self, *args, **kwargs):
        super(DirectorTestClient, self).__init__(*args, **kwargs)


def _remove_keys(d, keys=KEYS_TO_REMOVE):
    if isinstance(keys, str):
        keys = [keys]
    if isinstance(d, dict):
        for key in set(keys):
            if key in d:
                del d[key]
        for k, v in d.items():
            _remove_keys(v, keys)
    elif isinstance(d, list):
        for i in d:
            _remove_keys(i, keys)
    return d


class DirectorResponse(Response):
    _KEYS_TO_REMOVE = KEYS_TO_REMOVE

    @property
    def json(self):
        return _remove_keys(self.get_json(), self._KEYS_TO_REMOVE)


@pytest.fixture(scope="module")
def app_module():
    app = create_app(str(Path(__file__).parent.resolve() / "workflows"))

    with app.app_context():
        db.create_all()

    app.response_class = DirectorResponse
    app.test_client_class = DirectorTestClient
    yield app

    with app.app_context():
        db.drop_all()

    return app_module


@pytest.fixture(scope="function")
def app(app_module):
    with app_module.app_context():

        # Clean the relational database
        meta = db.metadata
        for table in reversed(meta.sorted_tables):
            db.session.execute(table.delete())
        db.session.commit()

    return app_module


@pytest.fixture
def client(app):
    return app.test_client()


@pytest.fixture(scope="function")
def enable_auth(app):
    with app.app_context():
        app.config["AUTH_ENABLED"] = True

    return app


@pytest.fixture(scope="function")
def no_worker(monkeypatch):
    monkeypatch.setattr(_chain, "apply_async", lambda x: None)


@pytest.fixture(autouse=True)
def skip_no_worker(app, request):
    """Skip a test if a Celery worker is not running"""
    if request.node.get_closest_marker("skip_no_worker"):
        with app.app_context():
            inspect = cel.control.inspect()
            if not inspect.ping():
                pytest.skip("Celery worker not running")


@pytest.fixture(scope="function")
def create_builder(app):
    def _create_builder(project, name, payload, periodic=False, keys=KEYS_TO_REMOVE):
        with app.app_context():
            obj = Workflow(
                project=project, name=name, payload=payload, periodic=periodic
            )
            obj.save()
            data = obj.to_dict()
            wf = WorkflowBuilder(obj.id)
            wf.build()
        return _remove_keys(data), wf

    return _create_builder



import os

import click

from director import create_app


class DirectorContext(click.Context):
    def __init__(self):
        self.app = create_app()
        self.app.app_context().push()


context = DirectorContext
pass_ctx = click.make_pass_decorator(context, ensure=True)



from director.builder import WorkflowBuilder
from director.extensions import cel
from director.models.workflows import Workflow


@cel.task()
def execute(workflow, payload):
    project, name = workflow.split(".")
    c_obj = Workflow(project=project, name=name, payload=payload, periodic=True)
    c_obj.save()

    # Build the workflow and execute it
    workflow = WorkflowBuilder(c_obj.id)
    workflow.run()

    c_obj_dict = c_obj.to_dict()

    # Force commit before ending the function to ensure the ongoing transaction
    # does not end up in a "idle in transaction" state on PostgreSQL
    c_obj.commit()

    return c_obj_dict



from datetime import datetime, timedelta
from unittest.mock import patch

from director.extensions import db
from director.models.workflows import Workflow


DEFAULT_PAYLOAD = {"project": "example", "name": "WORKFLOW", "payload": {}}


def test_pong(client):
    resp = client.get("/api/ping")
    assert resp.status_code == 200
    assert resp.json == {"message": "pong"}


def test_post_workflow_without_payload(client, no_worker):
    resp = client.post("/api/workflows", json=DEFAULT_PAYLOAD)
    assert resp.status_code == 201
    assert resp.json == {
        "fullname": "example.WORKFLOW",
        "name": "WORKFLOW",
        "payload": {},
        "periodic": False,
        "project": "example",
        "status": "pending",
    }


def test_post_workflow_with_payload(client, no_worker):
    payload = {**DEFAULT_PAYLOAD}
    payload["payload"] = {"nested": {"foo": "bar"}}
    resp = client.post("/api/workflows", json=payload)
    assert resp.status_code == 201
    assert resp.json == {
        "fullname": "example.WORKFLOW",
        "name": "WORKFLOW",
        "payload": {"nested": {"foo": "bar"}},
        "periodic": False,
        "project": "example",
        "status": "pending",
    }


def test_post_workflow_required_props(client):
    resp = client.post("/api/workflows", json={})
    assert resp.status_code == 400
    assert resp.json["error"] == "Error validating against schema"
    assert len(resp.json["errors"]) == 3
    for prop in ["project", "name", "payload"]:
        assert f"{prop}" in " ".join(resp.json["errors"])

    resp = client.post("/api/workflows", json={"project": "example"})
    assert resp.status_code == 400
    assert len(resp.json["errors"]) == 2
    for prop in ["name", "payload"]:
        assert f"{prop}" in " ".join(resp.json["errors"])

    resp = client.post(
        "/api/workflows", json={"project": "example", "name": "WORKFLOW"}
    )
    assert resp.status_code == 400
    assert len(resp.json["errors"]) == 1
    assert "payload" in resp.json["errors"][0]


def test_post_workflow_unexpected_properties(client):
    payload = {**DEFAULT_PAYLOAD, "foo": "bar"}
    resp = client.post("/api/workflows", json=payload)
    assert resp.status_code == 400
    assert resp.json["error"] == "Error validating against schema"
    assert "'foo' was unexpected" in resp.json["errors"][0]


def test_relaunch_not_existing_workflow(client):
    resp = client.post(
        "/api/workflows/280d0c4065d04063a800a3b674562711/relaunch", json={}
    )
    assert resp.status_code == 404


def test_relaunch_workflow(client, no_worker):
    payload = {**DEFAULT_PAYLOAD}
    payload["payload"] = {"nested": {"foo": "bar"}}
    resp = client.post("/api/workflows", json=payload)
    with patch("tests.conftest.DirectorResponse._KEYS_TO_REMOVE", new=[]):
        workflow_id = resp.json["id"]
    assert resp.status_code == 201

    resp = client.post(f"/api/workflows/{workflow_id}/relaunch", json={})
    assert resp.status_code == 201
    assert resp.json == {
        "fullname": "example.WORKFLOW",
        "name": "WORKFLOW",
        "payload": {"nested": {"foo": "bar"}},
        "periodic": False,
        "project": "example",
        "status": "pending",
    }

    # workflows are the same
    resp = client.get("/api/workflows")
    assert len(resp.json) == 2
    del resp.json[0]["status"]
    del resp.json[1]["status"]
    assert resp.json[0] == resp.json[1]

    # tasks are the same
    resp = client.get("/api/workflows")
    with patch("tests.conftest.DirectorResponse._KEYS_TO_REMOVE", new=[]):
        workflows = resp.json
    tasks1 = [
        t["key"]
        for t in client.get(f"/api/workflows/{workflows[0]['id']}").json.get("tasks")
    ]
    tasks2 = [
        t["key"]
        for t in client.get(f"/api/workflows/{workflows[1]['id']}").json.get("tasks")
    ]
    assert tasks1 == tasks2


def test_not_found_workflows(client, no_worker):
    resp = client.post("/api/workflows", json=DEFAULT_PAYLOAD)
    assert resp.status_code == 201

    payload = {**DEFAULT_PAYLOAD}
    payload["name"] = "NOT_FOUND"
    resp = client.post("/api/workflows", json=payload)
    assert resp.status_code == 404
    assert resp.json == {"error": "Workflow example.NOT_FOUND not found"}

    payload = {**DEFAULT_PAYLOAD}
    payload["project"] = "NOT_FOUND"
    resp = client.post("/api/workflows", json=payload)
    assert resp.status_code == 404
    assert resp.json == {"error": "Workflow NOT_FOUND.WORKFLOW not found"}


def test_list_workflows(client, no_worker):
    resp = client.get("/api/workflows")
    assert resp.status_code == 200
    assert resp.json == []

    resp = client.post("/api/workflows", json=DEFAULT_PAYLOAD)
    assert resp.status_code == 201

    resp = client.get("/api/workflows")
    assert resp.json == [
        {
            "fullname": "example.WORKFLOW",
            "name": "WORKFLOW",
            "payload": {},
            "periodic": False,
            "project": "example",
            "status": "pending",
        }
    ]

    for i in range(9):
        client.post("/api/workflows", json=DEFAULT_PAYLOAD)
    resp = client.get("/api/workflows")
    assert len(resp.json) == 10


def test_list_paginated_workflows(app, client, no_worker):
    # SQLite db.Datetime() does not include microseconds, so
    # we need to fix dates to be able to sort workflows.
    now = datetime.now()
    with app.app_context():
        for i in range(20):
            date = now + timedelta(minutes=i)
            w = Workflow(
                created_at=date,
                updated_at=date,
                project="example",
                name="WORKFLOW",
                payload={"i": i + 1},
            )
            db.session.add(w)
        db.session.commit()

    # DIRECTOR_WORKFLOWS_PER_PAGE=15 in tests/workflows/.env
    resp = client.get("/api/workflows")
    assert len(resp.json) == 15

    # Last item
    resp = client.get("/api/workflows?per_page=1")
    assert len(resp.json) == 1
    assert resp.json[0]["payload"]["i"] == 20

    # Last X items
    resp = client.get("/api/workflows?per_page=4&page=1")
    assert len(resp.json) == 4
    assert [w["payload"]["i"] for w in resp.json] == [20, 19, 18, 17]

    # Last X items, second page
    resp = client.get("/api/workflows?per_page=4&page=2")
    assert len(resp.json) == 4
    assert [w["payload"]["i"] for w in resp.json] == [16, 15, 14, 13]

    # Remaining items
    resp = client.get("/api/workflows?per_page=3&page=7")
    assert len(resp.json) == 2
    assert [w["payload"]["i"] for w in resp.json] == [2, 1]

    # Out of range
    resp = client.get("/api/workflows?per_page=4&page=6")
    assert resp.status_code == 404


def test_get_workflow(client, no_worker):
    resp = client.post("/api/workflows", json=DEFAULT_PAYLOAD)
    with patch("tests.conftest.DirectorResponse._KEYS_TO_REMOVE", new=[]):
        workflow_id = resp.json["id"]

    resp = client.get(f"/api/workflows/{workflow_id}")
    assert resp.status_code == 200
    assert len(resp.json["tasks"]) == 1

    del resp.json["tasks"][0]["task"]
    assert resp.json == {
        "fullname": "example.WORKFLOW",
        "name": "WORKFLOW",
        "payload": {},
        "periodic": False,
        "project": "example",
        "status": "pending",
        "tasks": [
            {
                "key": "TASK_EXAMPLE",
                "previous": [],
                "result": None,
                "status": "pending",
            }
        ],
    }


@patch("tests.conftest.DirectorResponse._KEYS_TO_REMOVE", new=[])
def test_get_workflow_simple_chain(client, no_worker):
    payload = {**DEFAULT_PAYLOAD}
    payload["name"] = "SIMPLE_CHAIN"
    workflow_id = client.post("/api/workflows", json=payload).json["id"]

    resp = client.get(f"/api/workflows/{workflow_id}")
    assert resp.status_code == 200
    assert len(resp.json["tasks"]) == 3


def test_get_not_found_workflow(client):
    resp = client.get("/api/workflows/e0ee42d8-e00f-4557-8f75-781106dbbea4")
    assert resp.status_code == 404


def test_schema(client, no_worker):
    payload = {"project": "schemas", "name": "SIMPLE_SCHEMA", "payload": {}}
    resp = client.post("/api/workflows", json=payload)
    assert resp.status_code == 400
    assert resp.json == {
        "error": "Payload is not valid",
        "errors": ["'name' is a required property"],
    }

    payload["payload"]["price"] = "foo"
    resp = client.post("/api/workflows", json=payload)
    assert resp.status_code == 400
    assert resp.json["error"] == "Payload is not valid"
    assert sorted(resp.json["errors"]) == sorted(
        ["'name' is a required property", "'foo' is not of type 'number'"]
    )

    payload["payload"] = {"name": "foo", "price": 100}
    resp = client.post("/api/workflows", json=payload)
    assert resp.status_code == 201



from celery.utils.log import get_task_logger

from director.extensions import cel
from director.tasks.base import BaseTask
from director.models import StatusType
from director.models.workflows import Workflow


logger = get_task_logger(__name__)


@cel.task(name="celery.ping")
def ping():
    # type: () -> str
    """Simple task that just returns 'pong'."""
    return "pong"


@cel.task()
def start(workflow_id):
    logger.info(f"Opening the workflow {workflow_id}")
    workflow = Workflow.query.filter_by(id=workflow_id).first()
    workflow.status = StatusType.progress
    workflow.save()


@cel.task()
def end(workflow_id):
    logger.info(f"Closing the workflow {workflow_id}")
    workflow = Workflow.query.filter_by(id=workflow_id).first()

    if workflow.status != StatusType.error:
        workflow.status = StatusType.success
        workflow.save()



import base64
import hashlib
from pathlib import Path
from urllib.request import urlretrieve

import click

from director.context import pass_ctx

DEPENDENCIES = [
    (
        "https://cdn.jsdelivr.net/npm/vue@2.6.11/dist/vue.min.js",
        "sha256-ngFW3UnAN0Tnm76mDuu7uUtYEcG3G5H1+zioJw3t+68=",
        None,
    ),
    (
        "https://cdn.jsdelivr.net/npm/vue-router@3.3.4/dist/vue-router.min.js",
        "sha256-87g98o5+aw7/ExJ9tBjvH8zz46FJ7hiylPtNFcPvsSw=",
        None,
    ),
    (
        "https://cdn.jsdelivr.net/npm/vuex@3.1.2/dist/vuex.min.js",
        "sha256-LfE9mPMjeOg3dTn1sESY2Xvdbq7gAhONtkxacnr7FSA=",
        None,
    ),
    (
        "https://cdn.jsdelivr.net/npm/vuetify@2.3.8/dist/vuetify.min.js",
        "sha256-b4RV6u+xflpPubfyN5gdooQRpDjROUcaCCSWbKQNX9Y=",
        None,
    ),
    (
        "https://cdn.jsdelivr.net/npm/moment@2.24.0/min/moment.min.js",
        "sha256-4iQZ6BVL4qNKlQ27TExEhBN1HFPvAvAMbFavKKosSWQ=",
        None,
    ),
    (
        "https://cdn.jsdelivr.net/npm/axios@0.19.0/dist/axios.min.js",
        "sha256-S1J4GVHHDMiirir9qsXWc8ZWw74PHHafpsHp5PXtjTs=",
        None,
    ),
    (
        "https://cdn.jsdelivr.net/npm/d3@5.16.0/dist/d3.min.js",
        "sha256-Xb6SSzhH3wEPC4Vy3W70Lqh9Y3Du/3KxPqI2JHQSpTw=",
        None,
    ),
    (
        "https://cdn.jsdelivr.net/npm/dagre-d3@0.6.4/dist/dagre-d3.min.js",
        "sha256-dPm4TA8Y9PY5q5mmtWMkRGOCMHJDKy34ZrxdbBGA9cs=",
        None,
    ),
    (
        "https://cdn.jsdelivr.net/npm/vuetify@2.3.8/dist/vuetify.min.css",
        "sha256-aS2abSwj+AMqtAmTglCnsg9GkAyDUKSCFd4xNdLX/AA=",
        None,
    ),
    (
        "https://cdn.jsdelivr.net/npm/@mdi/font@5.5.55/css/materialdesignicons.min.css",
        "sha256-HCkcFMiRi/WMAXrgcUj/C5aoUrPtvHxpfbOIlwdsNtg=",
        "mdi",
    ),
    (
        "https://fonts.gstatic.com/s/roboto/v20/KFOkCnqEu92Fr1MmgVxFIzIXKMnyrYk.woff2",
        "sha256-zlHOyHhuN60+uGhmzvAy+DRFPdAFH1q7iQet5ST8m1E=",
        "fonts",
    ),
    (
        "https://cdn.jsdelivr.net/npm/@mdi/font@5.5.55/fonts/materialdesignicons-webfont.woff2",
        "sha256-6PFQMmcHK85n15RxAODS3QGyyNG04kPtO+pFnw+Ud/c=",
        "fonts",
    ),
]


def compute_sri_hash(filename, block_size=4096):
    """
    Compute a SRI-compliant hash with SHA256
    :param filename: path to the file
    :param block_size: size of the block for reading binary data
    :return: a base64-encoded hash
    """
    sha256_hash = hashlib.sha256()
    with open(filename, "rb") as f:
        for block in iter(lambda: f.read(block_size), b""):
            sha256_hash.update(block)
        b64_encoded_str = base64.b64encode(sha256_hash.digest())
        return b64_encoded_str.decode()


@click.command()
@pass_ctx
def dlassets(ctx):
    """Download the required static assets"""
    static_path = Path(ctx.app.config["STATIC_FOLDER"])

    if static_path == Path(ctx.app.config["DIRECTOR_HOME"]) / "static":
        Path(static_path).mkdir(parents=True, exist_ok=True)

    for (dep_url, dep_hash, dep_custom_path) in DEPENDENCIES:
        # Get the name of the output file
        _, dep_filename = dep_url.rsplit("/", 1)

        if not dep_custom_path:
            path = static_path / dep_filename
        else:
            path = static_path / dep_custom_path / dep_filename
            Path(static_path / dep_custom_path).mkdir(parents=True, exist_ok=True)

        click.echo(f"Downloading {dep_filename} in {path}")

        # Download the file from the CDN
        try:
            urlretrieve(dep_url, str(path))
        except Exception as err:
            click.echo(click.style(f"The command encountered an error {err}", fg="red"))
            raise click.Abort()

        # Remove the "shaXXX-" part
        _, hash_value = dep_hash.split("-", 1)

        assert compute_sri_hash(str(path)) == hash_value



import json

from sqlalchemy_utils import UUIDType
from sqlalchemy.types import PickleType

from director.extensions import db
from director.exceptions import UserNotFound
from director.models import BaseModel, StatusType
from director.models.utils import JSONBType


class User(BaseModel):
    __tablename__ = "users"

    username = db.Column(db.String(255), unique=True, nullable=False)
    password = db.Column(db.String(255), nullable=False)

    def __repr__(self):
        return f"<User {self.username}>"

    def update(self):
        user = self.query.filter_by(username=self.username).first()
        if not user:
            raise UserNotFound(f"User {self.username} not found")

        user.password = self.password

        self.commit()

    def delete(self):
        db.session.delete(self)

        self.commit()

    def to_dict(self):
        d = super().to_dict()
        d.update({"username": self.username, "password": self.password})
        return d



import enum
import uuid

from sqlalchemy_utils import UUIDType

from director.extensions import db


def get_uuid():
    return str(uuid.uuid4())


class StatusType(enum.Enum):
    pending = "pending"
    progress = "progress"
    success = "success"
    error = "error"
    canceled = "canceled"


class BaseModel(db.Model):
    __abstract__ = True

    id = db.Column(
        UUIDType(binary=False), primary_key=True, nullable=False, default=get_uuid
    )
    created_at = db.Column(
        db.DateTime(timezone=True), default=db.func.now(), nullable=False, index=True
    )
    updated_at = db.Column(
        db.DateTime(timezone=True),
        default=db.func.now(),
        onupdate=db.func.now(),
        nullable=False,
    )

    def commit(self):
        try:
            db.session.commit()
        except Exception:
            db.session.rollback()
            raise

    def save(self):
        db.session.add(self)
        self.commit()

    def to_dict(self):
        return {
            "id": self.id,
            "created": self.created_at.isoformat(),
            "updated": self.updated_at.isoformat(),
        }



class WorkflowNotFound(Exception):
    pass


class SchemaNotFound(Exception):
    pass


class SchemaNotValid(Exception):
    pass


class WorkflowSyntaxError(Exception):
    pass


class UserNotFound(Exception):
    pass



import os
import click


@click.command(context_settings=dict(ignore_unknown_options=True))
@click.option("--dev", "dev_mode", default=False, is_flag=True, type=bool)
@click.argument("web_args", nargs=-1, type=click.UNPROCESSED)
def webserver(dev_mode, web_args):
    """Start the webserver instance"""
    if dev_mode:
        os.environ["FLASK_APP"] = f"director._auto:app"
        os.environ["FLASK_DEBUG"] = "1"
        args = [
            "flask",
            "run",
            "--port",
            "8000",
        ]
        args += list(web_args)
        os.execvpe(args[0], args, os.environ)
    else:
        args = [
            "gunicorn",
        ]
        args += list(web_args)
        args.append("director._auto:app")
        os.execvp(args[0], args)



from werkzeug.security import generate_password_hash, check_password_hash

from director.models.users import User


def test_crud(app, client):
    # Create user
    user = User(username="test_user", password=generate_password_hash("root_test1"))
    user.save()

    # Read user
    with app.app_context():
        user = User.query.filter_by(username="test_user").first()
    assert user.username == "test_user"
    assert check_password_hash(user.password, "root_test1")

    # Update user
    user.password = generate_password_hash("root_test2")
    user.update()
    user.save()

    with app.app_context():
        user = User.query.filter_by(username="test_user").first()
    assert check_password_hash(user.password, "root_test2")

    # Delete user
    user.delete()

    with app.app_context():
        user = User.query.filter_by(username="test_user").first()
    assert user is None



"""Initial migration

Revision ID: 30d6f6636351
Revises: 
Create Date: 2020-02-07 18:34:41.680883

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy_utils.types import UUIDType

from director.models.utils import JSONBType


# revision identifiers, used by Alembic.
revision = "30d6f6636351"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "workflows",
        sa.Column("id", UUIDType(binary=False), nullable=False),
        sa.Column("created_at", sa.DateTime(timezone=True), nullable=False),
        sa.Column("updated_at", sa.DateTime(timezone=True), nullable=False),
        sa.Column("name", sa.String(255), nullable=False),
        sa.Column("project", sa.String(255), nullable=False),
        sa.Column(
            "status",
            sa.Enum(
                "pending", "progress", "success", "error", "canceled", name="statustype"
            ),
            nullable=False,
        ),
        sa.Column("payload", JSONBType(), nullable=True),
        sa.Column("periodic", sa.Boolean(), nullable=True),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_workflows")),
    )
    op.create_index(
        op.f("ix_workflows_created_at"), "workflows", ["created_at"], unique=False
    )

    op.create_table(
        "tasks",
        sa.Column("id", UUIDType(binary=False), nullable=False),
        sa.Column("created_at", sa.DateTime(timezone=True), nullable=False),
        sa.Column("updated_at", sa.DateTime(timezone=True), nullable=False),
        sa.Column("key", sa.String(255), nullable=False),
        sa.Column(
            "status",
            sa.Enum(
                "pending", "progress", "success", "error", "canceled", name="statustype"
            ),
            nullable=False,
        ),
        sa.Column("previous", JSONBType(), nullable=True),
        sa.Column("workflow_id", UUIDType(binary=False), nullable=False),
        sa.ForeignKeyConstraint(
            ["workflow_id"],
            ["workflows.id"],
            name=op.f("fk_tasks_workflow_id_workflows"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_tasks")),
    )
    op.create_index(op.f("ix_tasks_created_at"), "tasks", ["created_at"], unique=False)


def downgrade():
    op.drop_index(op.f("ix_tasks_created_at"), table_name="tasks")
    op.drop_table("tasks")
    op.drop_index(op.f("ix_workflows_created_at"), table_name="workflows")
    op.drop_table("workflows")
    sa.Enum(name="statustype").drop(op.get_bind())



from flask import jsonify, Blueprint
from flask_json_schema import JsonValidationError

from director.utils import validate, format_schema_errors


# Main application
api_bp = Blueprint("api", __name__, url_prefix="/api")


@api_bp.errorhandler(JsonValidationError)
def schema_exception_handler(e):
    return jsonify(format_schema_errors(e)), 400


@api_bp.route("/ping")
def ping():
    return jsonify({"message": "pong"})



from celery import Task as _Task
from celery.signals import after_task_publish, task_prerun, task_postrun
from celery.utils.log import get_task_logger

from director.extensions import cel, db
from director.models import StatusType
from director.models.workflows import Workflow
from director.models.tasks import Task


logger = get_task_logger(__name__)


@task_prerun.connect
def director_prerun(task_id, task, *args, **kwargs):
    if task.name.startswith("director.tasks"):
        return

    with cel.app.app_context():
        task = Task.query.filter_by(id=task_id).first()
        task.status = StatusType.progress
        task.save()


@task_postrun.connect
def close_session(*args, **kwargs):
    # Flask SQLAlchemy will automatically create new sessions for you from
    # a scoped session factory, given that we are maintaining the same app
    # context, this ensures tasks have a fresh session (e.g. session errors
    # won't propagate across tasks)
    db.session.remove()


class BaseTask(_Task):
    def on_failure(self, exc, task_id, args, kwargs, einfo):
        task = Task.query.filter_by(id=task_id).first()
        task.status = StatusType.error
        task.result = {"exception": str(exc), "traceback": einfo.traceback}
        task.workflow.status = StatusType.error
        task.save()

        logger.info(f"Task {task_id} is now in error")
        super(BaseTask, self).on_failure(exc, task_id, args, kwargs, einfo)

    def on_success(self, retval, task_id, args, kwargs):
        task = Task.query.filter_by(id=task_id).first()
        task.status = StatusType.success
        task.result = retval
        task.save()

        logger.info(f"Task {task_id} is now in success")
        super(BaseTask, self).on_success(retval, task_id, args, kwargs)



from celery import chain, group
from celery.utils import uuid

from director.exceptions import WorkflowSyntaxError
from director.extensions import cel, cel_workflows
from director.models import StatusType
from director.models.tasks import Task
from director.models.workflows import Workflow
from director.tasks.workflows import start, end


class WorkflowBuilder(object):
    def __init__(self, workflow_id):
        self.workflow_id = workflow_id
        self._workflow = None

        self.queue = cel_workflows.get_queue(str(self.workflow))
        self.tasks = cel_workflows.get_tasks(str(self.workflow))
        self.canvas = []

        # Pointer to the previous task(s)
        self.previous = []

    @property
    def workflow(self):
        if not self._workflow:
            self._workflow = Workflow.query.filter_by(id=self.workflow_id).first()
        return self._workflow

    def new_task(self, task_name, single=True):
        task_id = uuid()

        # We create the Celery task specifying its UID
        signature = cel.tasks.get(task_name).subtask(
            kwargs={"workflow_id": self.workflow_id, "payload": self.workflow.payload},
            queue=self.queue,
            task_id=task_id,
        )

        # Director task has the same UID
        task = Task(
            id=task_id,
            key=task_name,
            previous=self.previous,
            workflow_id=self.workflow.id,
            status=StatusType.pending,
        )
        task.save()

        if single:
            self.previous = [signature.id]

        return signature

    def parse(self, tasks):
        canvas = []

        for task in tasks:
            if type(task) is str:
                signature = self.new_task(task)
                canvas.append(signature)
            elif type(task) is dict:
                name = list(task)[0]
                if "type" not in task[name] and task[name]["type"] != "group":
                    raise WorkflowSyntaxError()

                sub_canvas_tasks = [
                    self.new_task(t, single=False) for t in task[name]["tasks"]
                ]

                sub_canvas = group(*sub_canvas_tasks, task_id=uuid())
                canvas.append(sub_canvas)
                self.previous = [s.id for s in sub_canvas_tasks]
            else:
                raise WorkflowSyntaxError()

        return canvas

    def build(self):
        self.canvas = self.parse(self.tasks)
        self.canvas.insert(0, start.si(self.workflow.id).set(queue=self.queue))
        self.canvas.append(end.si(self.workflow.id).set(queue=self.queue))

    def run(self):
        if not self.canvas:
            self.build()

        canvas = chain(*self.canvas, task_id=uuid())

        try:
            return canvas.apply_async()
        except Exception as e:
            self.workflow.status = StatusType.error
            self.workflow.save()
            raise e



from pathlib import Path

import click


ENV_TEMPLATE = """# Auto-generated, please adjust.
# ---------- Database ---------- 
DIRECTOR_DATABASE_URI="sqlite:///{{director_home}}/director.db"
DIRECTOR_DATABASE_POOL_RECYCLE=-1


# ---------- Celery ----------
DIRECTOR_BROKER_URI="redis://127.0.0.1:6379/0"
DIRECTOR_RESULT_BACKEND_URI="redis://127.0.0.1:6379/1"


# ---------- Frontend ---------- 
DIRECTOR_API_URL="http://127.0.0.1:8000/api"
DIRECTOR_FLOWER_URL="http://127.0.0.1:5555"
DIRECTOR_ENABLE_DARK_THEME=false
DIRECTOR_ENABLE_HISTORY_MODE=false
DIRECTOR_REFRESH_INTERVAL=30000

# ---------- API ----------
DIRECTOR_WORKFLOWS_PER_PAGE=1000
DIRECTOR_AUTH_ENABLED = false

# These settings are designed to be used with the "director dlassets" command,
# the DIRECTOR_STATIC_FOLDER will be used if you set DIRECTOR_ENABLE_CDN to false.
DIRECTOR_ENABLE_CDN=true
DIRECTOR_STATIC_FOLDER=${DIRECTOR_HOME}/static

# ---------- Sentry ----------
DIRECTOR_SENTRY_DSN=""
"""


WORKFLOWS_YML_TEMPLATE = """---
# Simple ETL example
#
# +-----------+      +-------------+      +--------+
# |  EXTRACT  +----->+  TRANSFORM  +----->+  LOAD  |
# +-----------+      +-------------+      +--------+
#
example.ETL:
  tasks:
    - EXTRACT
    - TRANSFORM
    - LOAD

# Group of tasks example
#
# +----------+       +----------+
# |  RANDOM  |       |  RANDOM  |
# +----+-----+       +-----+----+
#      |     +-------+     |
#      +---->+  ADD  <-----+
#            +-------+
#
example.RANDOMS:
  tasks:
    - GROUP_RANDOMS:
        type: group
        tasks:
          - RANDOM
          - RANDOM
    - ADD
"""


ETL_PY_TEMPLATE = """from director import task


@task(name="EXTRACT")
def extract(*args, **kwargs):
    print("Extracting data")


@task(name="TRANSFORM")
def transform(*args, **kwargs):
    print("Transforming data")


@task(name="LOAD")
def load(*args, **kwargs):
    print("Loading data")
"""


GROUP_PY_TEMPLATE = """import random
from director import task


@task(name="RANDOM")
def generate_random(*args, **kwargs):
    payload = kwargs["payload"]
    return random.randint(payload.get("start", 0), payload.get("end", 10))


@task(name="ADD")
def add_randoms(*args, **kwargs):
    return sum(args[0])
"""


@click.command()
@click.argument("path")
def init(path):
    """
    Create a new project
    """
    user_project_path = Path(path).resolve()
    user_project_path.mkdir(parents=True, exist_ok=True)
    (user_project_path / "tasks").mkdir(exist_ok=True)

    with open(user_project_path / ".env", "w", encoding="utf-8") as f:
        f.write(ENV_TEMPLATE.replace("{{director_home}}", str(user_project_path)))

    with open(user_project_path / "workflows.yml", "w", encoding="utf-8") as f:
        f.write(WORKFLOWS_YML_TEMPLATE)

    with open(user_project_path / "tasks" / "etl.py", "w", encoding="utf-8") as f:
        f.write(ETL_PY_TEMPLATE)

    with open(user_project_path / "tasks" / "group.py", "w", encoding="utf-8") as f:
        f.write(GROUP_PY_TEMPLATE)

    click.echo(f"[*] Project created in {user_project_path}")
    click.echo("[*] Do not forget to initialize the database")
    click.echo("You can now export the DIRECTOR_HOME environment variable")



import click
import json
from json.decoder import JSONDecodeError
from terminaltables import AsciiTable

from flask_json_schema import JsonValidationError

from director.builder import WorkflowBuilder
from director.context import pass_ctx
from director.exceptions import WorkflowNotFound
from director.extensions import cel_workflows
from director.models.workflows import Workflow
from director.utils import validate, format_schema_errors


def tasks_to_ascii(tasks):
    tasks_str = ""
    # Wrap the tasks list
    for task in tasks:
        if type(task) == dict:
            group_name = list(task.keys())[0]
            tasks_str += f"Group {group_name}:\n"
            for task_name in task[group_name].get("tasks", []):
                tasks_str += f" └ {task_name}\n"
        else:
            tasks_str += f"{task}\n"

    # Just remove the last newline
    if tasks_str:
        tasks_str = tasks_str[:-1]

    return tasks_str


@click.group()
def workflow():
    """Manage the workflows"""


@workflow.command(name="list")
@pass_ctx
def list_workflow(ctx):
    """List the workflows"""
    workflows = {
        k: v
        for k, v in sorted(cel_workflows.workflows.items(), key=lambda item: item[0])
    }

    data = [[f"Workflows ({len(workflows)})", "Periodic", "Tasks"]]

    # Add a row for each workflow
    for name, conf in workflows.items():
        periodic = conf.get("periodic", {}).get("schedule", "--")
        tasks_str = tasks_to_ascii(conf["tasks"])
        data.append([name, periodic, tasks_str])

    table = AsciiTable(data)
    table.inner_row_border = True
    table.justify_columns[1] = "center"
    click.echo(table.table)


@workflow.command(name="show")
@click.argument("name")
@pass_ctx
def show_workflow(ctx, name):
    """Display details of a workflow"""
    try:
        _workflow = cel_workflows.get_by_name(name)
    except WorkflowNotFound as e:
        click.echo(f"Error: {e}")
        raise click.Abort()

    tasks_str = tasks_to_ascii(_workflow["tasks"])
    periodic = _workflow.get("periodic", {}).get("schedule", "--")
    payload = _workflow.get("periodic", {}).get("payload", {})

    # Construct the table
    table = AsciiTable(
        [
            ["Name", name],
            ["Tasks", tasks_str],
            ["Periodic", periodic],
            ["Default Payload", payload],
        ]
    )
    table.inner_heading_row_border = False
    table.inner_row_border = True
    click.echo(table.table)


@workflow.command(name="run")
@click.argument("fullname")
@click.argument("payload", required=False, default="{}")
@pass_ctx
def run_workflow(ctx, fullname, payload):
    """Execute a workflow"""
    try:
        wf = cel_workflows.get_by_name(fullname)
        payload = json.loads(payload)

        if "schema" in wf:
            try:
                validate(payload, wf["schema"])
            except JsonValidationError as e:
                result = format_schema_errors(e)

                click.echo(f"Error: {result['error']}")
                for err in result["errors"]:
                    click.echo(f"- {err}")
                raise click.Abort()

    except WorkflowNotFound as e:
        click.echo(f"Error: {e}")
        raise click.Abort()
    except (JSONDecodeError) as e:
        click.echo(f"Error in the payload : {e}")
        raise click.Abort()

    # Create the workflow object
    project, name = fullname.split(".")
    obj = Workflow(project=project, name=name, payload=payload)
    obj.save()

    # Build the canvas and execute it
    _workflow = WorkflowBuilder(obj.id)
    _workflow.run()



import os

import click


@click.command(context_settings=dict(ignore_unknown_options=True))
@click.argument("db_args", nargs=-1, type=click.UNPROCESSED)
def db(db_args):
    """Manage the database"""
    os.environ["FLASK_APP"] = "director._auto:app"
    args = ["flask", "db"] + list(db_args)
    os.execvpe(args[0], args, os.environ)



"""Force varchar 255

Revision ID: 2ac615d6850b
Revises: 063ff371f2da
Create Date: 2020-10-09 17:35:12.402690

"""
from alembic import op
import sqlalchemy as sa

from director.extensions import db


# revision identifiers, used by Alembic.
revision = "2ac615d6850b"
down_revision = "063ff371f2da"
branch_labels = None
depends_on = None

# Store the list of tables
tables = {t[0]: t[1] for t in db.metadata.tables.items()}


def upgrade():
    """
    This migration is only useful for an existing Celery Director instance.
    """
    with op.batch_alter_table("tasks", copy_from=tables["tasks"]) as batch_op:
        batch_op.alter_column(
            "key",
            existing_type=sa.VARCHAR(),
            type_=sa.String(length=255),
            existing_nullable=False,
        )

    with op.batch_alter_table("users", copy_from=tables["users"]) as batch_op:
        batch_op.alter_column(
            "password",
            existing_type=sa.VARCHAR(),
            type_=sa.String(length=255),
            existing_nullable=False,
        )
        batch_op.alter_column(
            "username",
            existing_type=sa.VARCHAR(),
            type_=sa.String(length=255),
            existing_nullable=False,
        )

    with op.batch_alter_table("workflows", copy_from=tables["workflows"]) as batch_op:
        batch_op.alter_column(
            "name",
            existing_type=sa.VARCHAR(),
            type_=sa.String(length=255),
            existing_nullable=False,
        )
        batch_op.alter_column(
            "project",
            existing_type=sa.VARCHAR(),
            type_=sa.String(length=255),
            existing_nullable=False,
        )


def downgrade():
    with op.batch_alter_table("workflows", copy_from=tables["workflows"]) as batch_op:
        batch_op.alter_column(
            "project",
            existing_type=sa.String(length=255),
            type_=sa.VARCHAR(),
            existing_nullable=False,
        )
        batch_op.alter_column(
            "name",
            existing_type=sa.String(length=255),
            type_=sa.VARCHAR(),
            existing_nullable=False,
        )

    with op.batch_alter_table("users", copy_from=tables["users"]) as batch_op:
        batch_op.alter_column(
            "username",
            existing_type=sa.String(length=255),
            type_=sa.VARCHAR(),
            existing_nullable=False,
        )
        batch_op.alter_column(
            "password",
            existing_type=sa.String(length=255),
            type_=sa.VARCHAR(),
            existing_nullable=False,
        )

    with op.batch_alter_table("tasks", copy_from=tables["tasks"]) as batch_op:
        batch_op.alter_column(
            "key",
            existing_type=sa.String(length=255),
            type_=sa.VARCHAR(),
            existing_nullable=False,
        )
